@@ -42,6 +42,7 @@ int ipc_helper_tcp_connection(void);
 int ipc_helper_closed_handle(void);
 int ipc_send_recv_helper(void);
 int ipc_helper_bind_twice(void);
+int ipc_helper_send_zero(void);
 int stdio_over_pipes_helper(void);
 int spawn_stdin_stdout(void);
 int spawn_tcp_server_helper(void);
@@ -104,6 +105,10 @@ static int maybe_run_test(int argc, char **argv) {
     return ipc_helper_bind_twice();
   }
 
+  if (strcmp(argv[1], "ipc_helper_send_zero") == 0) {
+    return ipc_helper_send_zero();
+  }
+
   if (strcmp(argv[1], "stdio_over_pipes_helper") == 0) {
     return stdio_over_pipes_helper();
   }
@@ -142,11 +147,11 @@ static int maybe_run_test(int argc, char **argv) {
   if (strcmp(argv[1], "spawn_helper5") == 0) {
     const char out[] = "fourth stdio!\n";
     notify_parent_process();
-#ifdef _WIN32
-    DWORD bytes;
-    WriteFile((HANDLE) _get_osfhandle(3), out, sizeof(out) - 1, &bytes, NULL);
-#else
     {
+#ifdef _WIN32
+      DWORD bytes;
+      WriteFile((HANDLE) _get_osfhandle(3), out, sizeof(out) - 1, &bytes, NULL);
+#else
       ssize_t r;
 
       do
@@ -154,8 +159,8 @@ static int maybe_run_test(int argc, char **argv) {
       while (r == -1 && errno == EINTR);
 
       fsync(3);
-    }
 #endif
+    }
     return 1;
   }
 
