@@ -27,10 +27,11 @@
 
 static uv_pipe_t channel;
 static uv_tcp_t tcp_server;
+static uv_tcp_t tcp_server2;
 static uv_tcp_t tcp_connection;
 
 static int exit_cb_called;
-static int read2_cb_called;
+static int read_cb_called;
 static int tcp_write_cb_called;
 static int tcp_read_cb_called;
 static int on_pipe_read_called;
@@ -38,13 +39,15 @@ static int local_conn_accepted;
 static int remote_conn_accepted;
 static int tcp_server_listening;
 static uv_write_t write_req;
-static uv_pipe_t channel;
-static uv_tcp_t tcp_server;
+static uv_write_t write_req2;
 static uv_write_t conn_notify_req;
 static int close_cb_called;
 static int connection_accepted;
 static int tcp_conn_read_cb_called;
 static int tcp_conn_write_cb_called;
+static int closed_handle_data_read;
+static int closed_handle_write;
+static int send_zero_write;
 
 typedef struct {
   uv_connect_t conn_req;
@@ -53,6 +56,16 @@ typedef struct {
 } tcp_conn;
 
 #define CONN_COUNT 100
+#define BACKLOG 128
+#define LARGE_SIZE 100000
+
+static uv_buf_t large_buf;
+static char buffer[LARGE_SIZE];
+static uv_write_t write_reqs[300];
+static int write_reqs_completed;
+
+static unsigned int write_until_data_queued(void);
+static void send_handle_and_close(void);
 
 
 static void close_server_conn_cb(uv_handle_t* handle) {
@@ -84,16 +97,22 @@ static void on_connection(uv_stream_t* server, int status) {
 }
 
 
-static void exit_cb(uv_process_t* process, int exit_status, int term_signal) {
+static void exit_cb(uv_process_t* process,
+                    int64_t exit_status,
+                    int term_signal) {
   printf("exit_cb\n");
   exit_cb_called++;
   ASSERT(exit_status == 0);
+  ASSERT(term_signal == 0);
   uv_close((uv_handle_t*)process, NULL);
 }
 
 
-static uv_buf_t on_alloc(uv_handle_t* handle, size_t suggested_size) {
-  return uv_buf_init(malloc(suggested_size), suggested_size);
+static void on_alloc(uv_handle_t* handle,
+                     size_t suggested_size,
+                     uv_buf_t* buf) {
+  buf->base = malloc(suggested_size);
+  buf->len = suggested_size;
 }
 
 
@@ -120,9 +139,12 @@ static void make_many_connections(void) {
     r = uv_tcp_init(uv_default_loop(), &conn->conn);
     ASSERT(r == 0);
 
-    addr = uv_ip4_addr("127.0.0.1", TEST_PORT);
+    ASSERT(0 == uv_ip4_addr("127.0.0.1", TEST_PORT, &addr));
 
-    r = uv_tcp_connect(&conn->conn_req, (uv_tcp_t*)&conn->conn, addr, connect_cb);
+    r = uv_tcp_connect(&conn->conn_req,
+                       (uv_tcp_t*) &conn->conn,
+                       (const struct sockaddr*) &addr,
+                       connect_cb);
     ASSERT(r == 0);
 
     conn->conn.data = conn;
@@ -130,34 +152,39 @@ static void make_many_connections(void) {
 }
 
 
-static void on_read(uv_pipe_t* pipe, ssize_t nread, uv_buf_t buf,
-    uv_handle_type pending) {
+static void on_read(uv_stream_t* handle,
+                    ssize_t nread,
+                    const uv_buf_t* buf) {
   int r;
+  uv_pipe_t* pipe;
+  uv_handle_type pending;
   uv_buf_t outbuf;
-  uv_err_t err;
+
+  pipe = (uv_pipe_t*) handle;
 
   if (nread == 0) {
     /* Everything OK, but nothing read. */
-    free(buf.base);
+    free(buf->base);
     return;
   }
 
   if (nread < 0) {
-    err = uv_last_error(pipe->loop);
-    if (err.code == UV_EOF) {
-      free(buf.base);
+    if (nread == UV_EOF) {
+      free(buf->base);
       return;
     }
 
-    printf("error recving on channel: %s\n", uv_strerror(err));
+    printf("error recving on channel: %s\n", uv_strerror(nread));
     abort();
   }
 
   fprintf(stderr, "got %d bytes\n", (int)nread);
 
+  pending = uv_pipe_pending_type(pipe);
   if (!tcp_server_listening) {
-    ASSERT(nread > 0 && buf.base && pending != UV_UNKNOWN_HANDLE);
-    read2_cb_called++;
+    ASSERT(1 == uv_pipe_pending_count(pipe));
+    ASSERT(nread > 0 && buf->base && pending != UV_UNKNOWN_HANDLE);
+    read_cb_called++;
 
     /* Accept the pending TCP server, and start listening on it. */
     ASSERT(pending == UV_TCP);
@@ -167,13 +194,13 @@ static void on_read(uv_pipe_t* pipe, ssize_t nread, uv_buf_t buf,
     r = uv_accept((uv_stream_t*)pipe, (uv_stream_t*)&tcp_server);
     ASSERT(r == 0);
 
-    r = uv_listen((uv_stream_t*)&tcp_server, 12, on_connection);
+    r = uv_listen((uv_stream_t*)&tcp_server, BACKLOG, on_connection);
     ASSERT(r == 0);
 
     tcp_server_listening = 1;
 
     /* Make sure that the expected data is correctly multiplexed. */
-    ASSERT(memcmp("hello\n", buf.base, nread) == 0);
+    ASSERT(memcmp("hello\n", buf->base, nread) == 0);
 
     outbuf = uv_buf_init("world\n", 6);
     r = uv_write(&write_req, (uv_stream_t*)pipe, &outbuf, 1, NULL);
@@ -181,16 +208,82 @@ static void on_read(uv_pipe_t* pipe, ssize_t nread, uv_buf_t buf,
 
     /* Create a bunch of connections to get both servers to accept. */
     make_many_connections();
-  } else if (memcmp("accepted_connection\n", buf.base, nread) == 0) {
+  } else if (memcmp("accepted_connection\n", buf->base, nread) == 0) {
     /* Remote server has accepted a connection.  Close the channel. */
+    ASSERT(0 == uv_pipe_pending_count(pipe));
     ASSERT(pending == UV_UNKNOWN_HANDLE);
     remote_conn_accepted = 1;
     uv_close((uv_handle_t*)&channel, NULL);
   }
 
-  free(buf.base);
+  free(buf->base);
 }
 
+#ifdef _WIN32
+static void on_read_listen_after_bound_twice(uv_stream_t* handle,
+                                             ssize_t nread,
+                                             const uv_buf_t* buf) {
+  int r;
+  uv_pipe_t* pipe;
+  uv_handle_type pending;
+
+  pipe = (uv_pipe_t*) handle;
+
+  if (nread == 0) {
+    /* Everything OK, but nothing read. */
+    free(buf->base);
+    return;
+  }
+
+  if (nread < 0) {
+    if (nread == UV_EOF) {
+      free(buf->base);
+      return;
+    }
+
+    printf("error recving on channel: %s\n", uv_strerror(nread));
+    abort();
+  }
+
+  fprintf(stderr, "got %d bytes\n", (int)nread);
+
+  ASSERT(uv_pipe_pending_count(pipe) > 0);
+  pending = uv_pipe_pending_type(pipe);
+  ASSERT(nread > 0 && buf->base && pending != UV_UNKNOWN_HANDLE);
+  read_cb_called++;
+
+  if (read_cb_called == 1) {
+    /* Accept the first TCP server, and start listening on it. */
+    ASSERT(pending == UV_TCP);
+    r = uv_tcp_init(uv_default_loop(), &tcp_server);
+    ASSERT(r == 0);
+
+    r = uv_accept((uv_stream_t*)pipe, (uv_stream_t*)&tcp_server);
+    ASSERT(r == 0);
+
+    r = uv_listen((uv_stream_t*)&tcp_server, BACKLOG, on_connection);
+    ASSERT(r == 0);
+  } else if (read_cb_called == 2) {
+    /* Accept the second TCP server, and start listening on it. */
+    ASSERT(pending == UV_TCP);
+    r = uv_tcp_init(uv_default_loop(), &tcp_server2);
+    ASSERT(r == 0);
+
+    r = uv_accept((uv_stream_t*)pipe, (uv_stream_t*)&tcp_server2);
+    ASSERT(r == 0);
+
+    r = uv_listen((uv_stream_t*)&tcp_server2, BACKLOG, on_connection);
+    ASSERT(r == UV_EADDRINUSE);
+
+    uv_close((uv_handle_t*)&tcp_server, NULL);
+    uv_close((uv_handle_t*)&tcp_server2, NULL);
+    ASSERT(0 == uv_pipe_pending_count(pipe));
+    uv_close((uv_handle_t*)&channel, NULL);
+  }
+
+  free(buf->base);
+}
+#endif
 
 void spawn_helper(uv_pipe_t* channel,
                   uv_process_t* process,
@@ -200,7 +293,7 @@ void spawn_helper(uv_pipe_t* channel,
   char exepath[1024];
   char* args[3];
   int r;
-  uv_stdio_container_t stdio[1];
+  uv_stdio_container_t stdio[3];
 
   r = uv_pipe_init(uv_default_loop(), channel, 1);
   ASSERT(r == 0);
@@ -219,14 +312,17 @@ void spawn_helper(uv_pipe_t* channel,
   options.file = exepath;
   options.args = args;
   options.exit_cb = exit_cb;
-
   options.stdio = stdio;
-  options.stdio[0].flags = UV_CREATE_PIPE |
-    UV_READABLE_PIPE | UV_WRITABLE_PIPE;
-  options.stdio[0].data.stream = (uv_stream_t*)channel;
-  options.stdio_count = 1;
+  options.stdio_count = ARRAY_SIZE(stdio);
 
-  r = uv_spawn(uv_default_loop(), process, options);
+  stdio[0].flags = UV_CREATE_PIPE | UV_READABLE_PIPE | UV_WRITABLE_PIPE;
+  stdio[0].data.stream = (uv_stream_t*) channel;
+  stdio[1].flags = UV_INHERIT_FD;
+  stdio[1].data.fd = 1;
+  stdio[2].flags = UV_INHERIT_FD;
+  stdio[2].data.fd = 2;
+
+  r = uv_spawn(uv_default_loop(), process, &options);
   ASSERT(r == 0);
 }
 
@@ -238,19 +334,19 @@ static void on_tcp_write(uv_write_t* req, int status) {
 }
 
 
-static uv_buf_t on_read_alloc(uv_handle_t* handle, size_t suggested_size) {
-  uv_buf_t buf;
-  buf.base = (char*)malloc(suggested_size);
-  buf.len = suggested_size;
-  return buf;
+static void on_read_alloc(uv_handle_t* handle,
+                          size_t suggested_size,
+                          uv_buf_t* buf) {
+  buf->base = malloc(suggested_size);
+  buf->len = suggested_size;
 }
 
 
-static void on_tcp_read(uv_stream_t* tcp, ssize_t nread, uv_buf_t buf) {
+static void on_tcp_read(uv_stream_t* tcp, ssize_t nread, const uv_buf_t* buf) {
   ASSERT(nread > 0);
-  ASSERT(memcmp("hello again\n", buf.base, nread) == 0);
+  ASSERT(memcmp("hello again\n", buf->base, nread) == 0);
   ASSERT(tcp == (uv_stream_t*)&tcp_connection);
-  free(buf.base);
+  free(buf->base);
 
   tcp_read_cb_called++;
 
@@ -259,44 +355,49 @@ static void on_tcp_read(uv_stream_t* tcp, ssize_t nread, uv_buf_t buf) {
 }
 
 
-static void on_read_connection(uv_pipe_t* pipe, ssize_t nread, uv_buf_t buf,
-    uv_handle_type pending) {
+static void on_read_connection(uv_stream_t* handle,
+                               ssize_t nread,
+                               const uv_buf_t* buf) {
   int r;
   uv_buf_t outbuf;
-  uv_err_t err;
+  uv_pipe_t* pipe;
+  uv_handle_type pending;
 
+  pipe = (uv_pipe_t*) handle;
   if (nread == 0) {
     /* Everything OK, but nothing read. */
-    free(buf.base);
+    free(buf->base);
     return;
   }
 
   if (nread < 0) {
-    err = uv_last_error(pipe->loop);
-    if (err.code == UV_EOF) {
-      free(buf.base);
+    if (nread == UV_EOF) {
+      free(buf->base);
       return;
     }
 
-    printf("error recving on channel: %s\n", uv_strerror(err));
+    printf("error recving on channel: %s\n", uv_strerror(nread));
     abort();
   }
 
   fprintf(stderr, "got %d bytes\n", (int)nread);
 
-  ASSERT(nread > 0 && buf.base && pending != UV_UNKNOWN_HANDLE);
-  read2_cb_called++;
+  ASSERT(1 == uv_pipe_pending_count(pipe));
+  pending = uv_pipe_pending_type(pipe);
+
+  ASSERT(nread > 0 && buf->base && pending != UV_UNKNOWN_HANDLE);
+  read_cb_called++;
 
   /* Accept the pending TCP connection */
   ASSERT(pending == UV_TCP);
   r = uv_tcp_init(uv_default_loop(), &tcp_connection);
   ASSERT(r == 0);
 
-  r = uv_accept((uv_stream_t*)pipe, (uv_stream_t*)&tcp_connection);
+  r = uv_accept(handle, (uv_stream_t*)&tcp_connection);
   ASSERT(r == 0);
 
   /* Make sure that the expected data is correctly multiplexed. */
-  ASSERT(memcmp("hello\n", buf.base, nread) == 0);
+  ASSERT(memcmp("hello\n", buf->base, nread) == 0);
 
   /* Write/read to/from the connection */
   outbuf = uv_buf_init("world\n", 6);
@@ -307,16 +408,44 @@ static void on_read_connection(uv_pipe_t* pipe, ssize_t nread, uv_buf_t buf,
   r = uv_read_start((uv_stream_t*)&tcp_connection, on_read_alloc, on_tcp_read);
   ASSERT(r == 0);
 
-  free(buf.base);
+  free(buf->base);
 }
 
 
-static int run_ipc_test(const char* helper, uv_read2_cb read_cb) {
+#ifndef _WIN32
+static void on_read_closed_handle(uv_stream_t* handle,
+                                  ssize_t nread,
+                                  const uv_buf_t* buf) {
+  if (nread == 0 || nread == UV_EOF) {
+    free(buf->base);
+    return;
+  }
+
+  if (nread < 0) {
+    printf("error recving on channel: %s\n", uv_strerror(nread));
+    abort();
+  }
+
+  closed_handle_data_read += nread;
+  free(buf->base);
+}
+#endif
+
+
+static void on_read_send_zero(uv_stream_t* handle,
+                              ssize_t nread,
+                              const uv_buf_t* buf) {
+  ASSERT(nread == 0 || nread == UV_EOF);
+  free(buf->base);
+}
+
+
+static int run_ipc_test(const char* helper, uv_read_cb read_cb) {
   uv_process_t process;
   int r;
 
   spawn_helper(&channel, &process, helper);
-  uv_read2_start((uv_stream_t*)&channel, on_alloc, read_cb);
+  uv_read_start((uv_stream_t*)&channel, on_alloc, read_cb);
 
   r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
   ASSERT(r == 0);
@@ -327,45 +456,65 @@ static int run_ipc_test(const char* helper, uv_read2_cb read_cb) {
 
 
 TEST_IMPL(ipc_listen_before_write) {
+#if defined(NO_SEND_HANDLE_ON_PIPE)
+  RETURN_SKIP(NO_SEND_HANDLE_ON_PIPE);
+#endif
   int r = run_ipc_test("ipc_helper_listen_before_write", on_read);
   ASSERT(local_conn_accepted == 1);
   ASSERT(remote_conn_accepted == 1);
-  ASSERT(read2_cb_called == 1);
+  ASSERT(read_cb_called == 1);
   ASSERT(exit_cb_called == 1);
   return r;
 }
 
 
 TEST_IMPL(ipc_listen_after_write) {
+#if defined(NO_SEND_HANDLE_ON_PIPE)
+  RETURN_SKIP(NO_SEND_HANDLE_ON_PIPE);
+#endif
   int r = run_ipc_test("ipc_helper_listen_after_write", on_read);
   ASSERT(local_conn_accepted == 1);
   ASSERT(remote_conn_accepted == 1);
-  ASSERT(read2_cb_called == 1);
+  ASSERT(read_cb_called == 1);
   ASSERT(exit_cb_called == 1);
   return r;
 }
 
 
 TEST_IMPL(ipc_tcp_connection) {
+#if defined(NO_SEND_HANDLE_ON_PIPE)
+  RETURN_SKIP(NO_SEND_HANDLE_ON_PIPE);
+#endif
   int r = run_ipc_test("ipc_helper_tcp_connection", on_read_connection);
-  ASSERT(read2_cb_called == 1);
+  ASSERT(read_cb_called == 1);
   ASSERT(tcp_write_cb_called == 1);
   ASSERT(tcp_read_cb_called == 1);
   ASSERT(exit_cb_called == 1);
   return r;
 }
 
+#ifndef _WIN32
+TEST_IMPL(ipc_closed_handle) {
+  int r;
+  r = run_ipc_test("ipc_helper_closed_handle", on_read_closed_handle);
+  ASSERT(r == 0);
+  return 0;
+}
+#endif
+
 
 #ifdef _WIN32
 TEST_IMPL(listen_with_simultaneous_accepts) {
   uv_tcp_t server;
   int r;
-  struct sockaddr_in addr = uv_ip4_addr("0.0.0.0", TEST_PORT);
+  struct sockaddr_in addr;
+
+  ASSERT(0 == uv_ip4_addr("0.0.0.0", TEST_PORT, &addr));
 
   r = uv_tcp_init(uv_default_loop(), &server);
   ASSERT(r == 0);
 
-  r = uv_tcp_bind(&server, addr);
+  r = uv_tcp_bind(&server, (const struct sockaddr*) &addr, 0);
   ASSERT(r == 0);
 
   r = uv_tcp_simultaneous_accepts(&server, 1);
@@ -383,12 +532,14 @@ TEST_IMPL(listen_with_simultaneous_accepts) {
 TEST_IMPL(listen_no_simultaneous_accepts) {
   uv_tcp_t server;
   int r;
-  struct sockaddr_in addr = uv_ip4_addr("0.0.0.0", TEST_PORT);
+  struct sockaddr_in addr;
+
+  ASSERT(0 == uv_ip4_addr("0.0.0.0", TEST_PORT, &addr));
 
   r = uv_tcp_init(uv_default_loop(), &server);
   ASSERT(r == 0);
 
-  r = uv_tcp_bind(&server, addr);
+  r = uv_tcp_bind(&server, (const struct sockaddr*) &addr, 0);
   ASSERT(r == 0);
 
   r = uv_tcp_simultaneous_accepts(&server, 0);
@@ -401,7 +552,24 @@ TEST_IMPL(listen_no_simultaneous_accepts) {
   MAKE_VALGRIND_HAPPY();
   return 0;
 }
+
+TEST_IMPL(ipc_listen_after_bind_twice) {
+#if defined(NO_SEND_HANDLE_ON_PIPE)
+  RETURN_SKIP(NO_SEND_HANDLE_ON_PIPE);
 #endif
+  int r = run_ipc_test("ipc_helper_bind_twice", on_read_listen_after_bound_twice);
+  ASSERT(read_cb_called == 2);
+  ASSERT(exit_cb_called == 1);
+  return r;
+}
+#endif
+
+TEST_IMPL(ipc_send_zero) {
+  int r;
+  r = run_ipc_test("ipc_helper_send_zero", on_read_send_zero);
+  ASSERT(r == 0);
+  return 0;
+}
 
 
 /* Everything here runs in a child process. */
@@ -429,25 +597,48 @@ static void tcp_connection_write_cb(uv_write_t* req, int status) {
 }
 
 
-static void on_tcp_child_process_read(uv_stream_t* tcp, ssize_t nread, uv_buf_t buf) {
+static void closed_handle_large_write_cb(uv_write_t* req, int status) {
+  ASSERT(status == 0);
+  ASSERT(closed_handle_data_read = LARGE_SIZE);
+  if (++write_reqs_completed == ARRAY_SIZE(write_reqs)) {
+    write_reqs_completed = 0;
+    if (write_until_data_queued() > 0)
+      send_handle_and_close();
+  }
+}
+
+
+static void closed_handle_write_cb(uv_write_t* req, int status) {
+  ASSERT(status == UV_EBADF);
+  closed_handle_write = 1;
+}
+
+
+static void send_zero_write_cb(uv_write_t* req, int status) {
+  ASSERT(status == 0);
+  send_zero_write++;
+}
+
+static void on_tcp_child_process_read(uv_stream_t* tcp,
+                                      ssize_t nread,
+                                      const uv_buf_t* buf) {
   uv_buf_t outbuf;
   int r;
 
   if (nread < 0) {
-    if (uv_last_error(tcp->loop).code == UV_EOF) {
-      free(buf.base);
+    if (nread == UV_EOF) {
+      free(buf->base);
       return;
     }
 
-    printf("error recving on tcp connection: %s\n",
-      uv_strerror(uv_last_error(tcp->loop)));
+    printf("error recving on tcp connection: %s\n", uv_strerror(nread));
     abort();
   }
 
   ASSERT(nread > 0);
-  ASSERT(memcmp("world\n", buf.base, nread) == 0);
+  ASSERT(memcmp("world\n", buf->base, nread) == 0);
   on_pipe_read_called++;
-  free(buf.base);
+  free(buf->base);
 
   /* Write to the socket */
   outbuf = uv_buf_init("hello again\n", 12);
@@ -520,7 +711,9 @@ static void ipc_on_connection_tcp_conn(uv_stream_t* server, int status) {
     (uv_stream_t*)conn, NULL);
   ASSERT(r == 0);
 
-  r = uv_read_start((uv_stream_t*)conn, on_read_alloc, on_tcp_child_process_read);
+  r = uv_read_start((uv_stream_t*) conn,
+                    on_read_alloc,
+                    on_tcp_child_process_read);
   ASSERT(r == 0);
 
   uv_close((uv_handle_t*)conn, close_cb);
@@ -532,28 +725,29 @@ int ipc_helper(int listen_after_write) {
    * This is launched from test-ipc.c. stdin is a duplex channel that we
    * over which a handle will be transmitted.
    */
-
-  uv_write_t write_req;
+  struct sockaddr_in addr;
   int r;
   uv_buf_t buf;
 
+  ASSERT(0 == uv_ip4_addr("0.0.0.0", TEST_PORT, &addr));
+
   r = uv_pipe_init(uv_default_loop(), &channel, 1);
   ASSERT(r == 0);
 
   uv_pipe_open(&channel, 0);
 
-  ASSERT(uv_is_readable((uv_stream_t*) &channel));
-  ASSERT(uv_is_writable((uv_stream_t*) &channel));
-  ASSERT(!uv_is_closing((uv_handle_t*) &channel));
+  ASSERT(1 == uv_is_readable((uv_stream_t*) &channel));
+  ASSERT(1 == uv_is_writable((uv_stream_t*) &channel));
+  ASSERT(0 == uv_is_closing((uv_handle_t*) &channel));
 
   r = uv_tcp_init(uv_default_loop(), &tcp_server);
   ASSERT(r == 0);
 
-  r = uv_tcp_bind(&tcp_server, uv_ip4_addr("0.0.0.0", TEST_PORT));
+  r = uv_tcp_bind(&tcp_server, (const struct sockaddr*) &addr, 0);
   ASSERT(r == 0);
 
   if (!listen_after_write) {
-    r = uv_listen((uv_stream_t*)&tcp_server, 12, ipc_on_connection);
+    r = uv_listen((uv_stream_t*)&tcp_server, BACKLOG, ipc_on_connection);
     ASSERT(r == 0);
   }
 
@@ -563,10 +757,11 @@ int ipc_helper(int listen_after_write) {
   ASSERT(r == 0);
 
   if (listen_after_write) {
-    r = uv_listen((uv_stream_t*)&tcp_server, 12, ipc_on_connection);
+    r = uv_listen((uv_stream_t*)&tcp_server, BACKLOG, ipc_on_connection);
     ASSERT(r == 0);
   }
 
+  notify_parent_process();
   r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
   ASSERT(r == 0);
 
@@ -580,7 +775,7 @@ int ipc_helper(int listen_after_write) {
 
 int ipc_helper_tcp_connection(void) {
   /*
-   * This is launched from test-ipc.c. stdin is a duplex channel that we
+   * This is launched from test-ipc.c. stdin is a duplex channel
    * over which a handle will be transmitted.
    */
 
@@ -592,25 +787,31 @@ int ipc_helper_tcp_connection(void) {
 
   uv_pipe_open(&channel, 0);
 
-  ASSERT(uv_is_readable((uv_stream_t*)&channel));
-  ASSERT(uv_is_writable((uv_stream_t*)&channel));
-  ASSERT(!uv_is_closing((uv_handle_t*)&channel));
+  ASSERT(1 == uv_is_readable((uv_stream_t*) &channel));
+  ASSERT(1 == uv_is_writable((uv_stream_t*) &channel));
+  ASSERT(0 == uv_is_closing((uv_handle_t*) &channel));
 
   r = uv_tcp_init(uv_default_loop(), &tcp_server);
   ASSERT(r == 0);
 
-  r = uv_tcp_bind(&tcp_server, uv_ip4_addr("0.0.0.0", TEST_PORT));
+  ASSERT(0 == uv_ip4_addr("0.0.0.0", TEST_PORT, &addr));
+
+  r = uv_tcp_bind(&tcp_server, (const struct sockaddr*) &addr, 0);
   ASSERT(r == 0);
 
-  r = uv_listen((uv_stream_t*)&tcp_server, 12, ipc_on_connection_tcp_conn);
+  r = uv_listen((uv_stream_t*)&tcp_server, BACKLOG, ipc_on_connection_tcp_conn);
   ASSERT(r == 0);
 
   /* Make a connection to the server */
   r = uv_tcp_init(uv_default_loop(), &conn.conn);
   ASSERT(r == 0);
 
-  addr = uv_ip4_addr("127.0.0.1", TEST_PORT);
-  r = uv_tcp_connect(&conn.conn_req, (uv_tcp_t*)&conn.conn, addr, connect_child_process_cb);
+  ASSERT(0 == uv_ip4_addr("127.0.0.1", TEST_PORT, &addr));
+
+  r = uv_tcp_connect(&conn.conn_req,
+                     (uv_tcp_t*) &conn.conn,
+                     (const struct sockaddr*) &addr,
+                     connect_child_process_cb);
   ASSERT(r == 0);
 
   r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
@@ -623,3 +824,151 @@ int ipc_helper_tcp_connection(void) {
   MAKE_VALGRIND_HAPPY();
   return 0;
 }
+
+static unsigned int write_until_data_queued() {
+  unsigned int i;
+  int r;
+
+  i = 0;
+  do {
+    r = uv_write(&write_reqs[i],
+                 (uv_stream_t*)&channel,
+                 &large_buf,
+                 1,
+                 closed_handle_large_write_cb);
+    ASSERT(r == 0);
+    i++;
+  } while (((uv_stream_t*)&channel)->write_queue_size == 0 &&
+           i < ARRAY_SIZE(write_reqs));
+
+  return ((uv_stream_t*)&channel)->write_queue_size;
+}
+
+static void send_handle_and_close() {
+  int r;
+  struct sockaddr_in addr;
+
+  r = uv_tcp_init(uv_default_loop(), &tcp_server);
+  ASSERT(r == 0);
+
+  ASSERT(0 == uv_ip4_addr("0.0.0.0", TEST_PORT, &addr));
+
+  r = uv_tcp_bind(&tcp_server, (const struct sockaddr*) &addr, 0);
+  ASSERT(r == 0);
+
+  r = uv_write2(&write_req,
+                (uv_stream_t*)&channel,
+                &large_buf,
+                1,
+                (uv_stream_t*)&tcp_server,
+                closed_handle_write_cb);
+  ASSERT(r == 0);
+
+  uv_close((uv_handle_t*)&tcp_server, NULL);
+}
+
+int ipc_helper_closed_handle(void) {
+  int r;
+
+  memset(buffer, '.', LARGE_SIZE);
+  large_buf = uv_buf_init(buffer, LARGE_SIZE);
+
+  r = uv_pipe_init(uv_default_loop(), &channel, 1);
+  ASSERT(r == 0);
+
+  uv_pipe_open(&channel, 0);
+
+  ASSERT(1 == uv_is_readable((uv_stream_t*) &channel));
+  ASSERT(1 == uv_is_writable((uv_stream_t*) &channel));
+  ASSERT(0 == uv_is_closing((uv_handle_t*) &channel));
+
+  if (write_until_data_queued() > 0)
+    send_handle_and_close();
+
+  r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
+  ASSERT(r == 0);
+
+  ASSERT(closed_handle_write == 1);
+
+  MAKE_VALGRIND_HAPPY();
+  return 0;
+}
+
+
+int ipc_helper_bind_twice(void) {
+  /*
+   * This is launched from test-ipc.c. stdin is a duplex channel
+   * over which two handles will be transmitted.
+   */
+  struct sockaddr_in addr;
+  int r;
+  uv_buf_t buf;
+
+  ASSERT(0 == uv_ip4_addr("0.0.0.0", TEST_PORT, &addr));
+
+  r = uv_pipe_init(uv_default_loop(), &channel, 1);
+  ASSERT(r == 0);
+
+  uv_pipe_open(&channel, 0);
+
+  ASSERT(1 == uv_is_readable((uv_stream_t*) &channel));
+  ASSERT(1 == uv_is_writable((uv_stream_t*) &channel));
+  ASSERT(0 == uv_is_closing((uv_handle_t*) &channel));
+
+  buf = uv_buf_init("hello\n", 6);
+
+  r = uv_tcp_init(uv_default_loop(), &tcp_server);
+  ASSERT(r == 0);
+  r = uv_tcp_init(uv_default_loop(), &tcp_server2);
+  ASSERT(r == 0);
+
+  r = uv_tcp_bind(&tcp_server, (const struct sockaddr*) &addr, 0);
+  ASSERT(r == 0);
+  r = uv_tcp_bind(&tcp_server2, (const struct sockaddr*) &addr, 0);
+  ASSERT(r == 0);
+
+  r = uv_write2(&write_req, (uv_stream_t*)&channel, &buf, 1,
+                (uv_stream_t*)&tcp_server, NULL);
+  ASSERT(r == 0);
+  r = uv_write2(&write_req2, (uv_stream_t*)&channel, &buf, 1,
+                (uv_stream_t*)&tcp_server2, NULL);
+  ASSERT(r == 0);
+
+  r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
+  ASSERT(r == 0);
+
+  MAKE_VALGRIND_HAPPY();
+  return 0;
+}
+
+int ipc_helper_send_zero(void) {
+  int r;
+  uv_buf_t zero_buf;
+
+  zero_buf = uv_buf_init(0, 0);
+
+  r = uv_pipe_init(uv_default_loop(), &channel, 0);
+  ASSERT(r == 0);
+
+  uv_pipe_open(&channel, 0);
+
+  ASSERT(1 == uv_is_readable((uv_stream_t*) &channel));
+  ASSERT(1 == uv_is_writable((uv_stream_t*) &channel));
+  ASSERT(0 == uv_is_closing((uv_handle_t*) &channel));
+
+  r = uv_write(&write_req,
+               (uv_stream_t*)&channel,
+               &zero_buf,
+               1,
+               send_zero_write_cb);
+
+  ASSERT(r == 0);
+
+  r = uv_run(uv_default_loop(), UV_RUN_DEFAULT);
+  ASSERT(r == 0);
+
+  ASSERT(send_zero_write == 1);
+
+  MAKE_VALGRIND_HAPPY();
+  return 0;
+}
\ No newline at end of file
